package de.mdelab.mlsdm.interpreter.callActions;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.util.EcoreUtil;

import de.mdelab.mlcallactions.ActivityCallAction;
import de.mdelab.mlcallactions.BinaryOperationAction;
import de.mdelab.mlcallactions.BinaryOperatorsEnum;
import de.mdelab.mlcallactions.CallAction;
import de.mdelab.mlcallactions.CallActionExpression;
import de.mdelab.mlcallactions.CallActionParameter;
import de.mdelab.mlcallactions.CloneAction;
import de.mdelab.mlcallactions.EObjectReferenceAction;
import de.mdelab.mlcallactions.EOperationCallAction;
import de.mdelab.mlcallactions.LiteralDeclarationAction;
import de.mdelab.mlcallactions.MethodCallAction;
import de.mdelab.mlcallactions.MlcallactionsPackage;
import de.mdelab.mlcallactions.NewObjectAction;
import de.mdelab.mlcallactions.NullValueAction;
import de.mdelab.mlcallactions.UnaryOperationAction;
import de.mdelab.mlcallactions.VariableDeclarationAction;
import de.mdelab.mlcallactions.VariableReferenceAction;
import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.interpreter.MLSDMInterpreter;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.SDMInterpreterConstants;
import de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreter;
import de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreterManager;
import de.mdelab.sdm.interpreter.core.notifications.NotificationEmitter;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.core.variables.VariablesScope;

/**
 * The CallActionInterpreter executes CallActions, an expression language
 * without a concrete textual syntax. Its main purpose is to allow calling Java
 * methods and other story diagrams from a story diagram.
 * 
 * @author Stephan Hildebrandt
 * 
 */
public class CallActionInterpreter extends ExpressionInterpreter<MLExpression, EClassifier>
{

	@Override
	public Variable<EClassifier> evaluateExpression(final MLExpression expression, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		assert expression instanceof CallActionExpression : "The CallActionInterpreter can only evaluate CallActionExpressions, but not '"
				+ expression.eClass() + "'.";

		Variable<EClassifier> variable = null;

		for (final CallAction callAction : ((CallActionExpression) expression).getCallActions())
		{
			final EClass callActionEClass = callAction.eClass();

			if (callActionEClass == MlcallactionsPackage.Literals.LITERAL_DECLARATION_ACTION)
			{
				variable = this.evaluate((LiteralDeclarationAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.ACTIVITY_CALL_ACTION)
			{
				variable = this.evaluate((ActivityCallAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.NEW_OBJECT_ACTION)
			{
				variable = this.evaluate((NewObjectAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_DECLARATION_ACTION)
			{
				variable = this.evaluate((VariableDeclarationAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.UNARY_OPERATION_ACTION)
			{
				variable = this.evaluate((UnaryOperationAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.BINARY_OPERATION_ACTION)
			{
				variable = this.evaluate((BinaryOperationAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.NULL_VALUE_ACTION)
			{
				variable = this.evaluate((NullValueAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.METHOD_CALL_ACTION)
			{
				variable = this.evaluate((MethodCallAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.EOPERATION_CALL_ACTION)
			{
				variable = this.evaluate((EOperationCallAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.VARIABLE_REFERENCE_ACTION)
			{
				variable = this.evaluate((VariableReferenceAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else if (callActionEClass == MlcallactionsPackage.Literals.EOBJECT_REFERENCE_ACTION)
			{
				variable = this.evaluate((EObjectReferenceAction) callAction, contextClassifier, contextInstance, variablesScope);
			}
			else
			{
				throw new UnsupportedOperationException();
			}
		}

		if (variable == null)
		{
			throw new UnsupportedOperationException();
		}
		else
		{
			return variable;
		}
	}

	private Variable<EClassifier> evaluate(final MLExpression valueExpression, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{

		if (valueExpression.eClass() == MlcallactionsPackage.Literals.CALL_ACTION_EXPRESSION)
		{
			return this.evaluateExpression(valueExpression, contextClassifier, contextInstance, variablesScope);
		}
		else
		{
			return this.getExpressionInterpreterManager().evaluateExpression(valueExpression, contextClassifier, contextInstance,
					variablesScope);
		}
	}

	protected Variable<EClassifier> evaluate(final VariableReferenceAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		final Variable<EClassifier> variable = variablesScope.getVariable(callAction.getName());

		if (variable != null)
		{
			assert variable.getClassifier() == callAction.getType() || variable.getClassifier() instanceof EClass
					&& callAction.getType() instanceof EClass
					&& ((EClass) callAction.getType()).isSuperTypeOf((EClass) variable.getClassifier());

			return variable;
		}
		else
		{
			throw new SDMException("The variable " + callAction.getName() + " is undefined.");
		}
	}

	protected Variable<EClassifier> evaluate(final MethodCallAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		final CallActionParameter[] parameters = callAction.getParameters().toArray(new CallActionParameter[] {});

		// Create array of parameter types and parameters
		final Class<?>[] parameterTypes = new Class<?>[parameters.length];
		final Object[] parameterObjects = new Object[parameters.length];

		try
		{
			for (int i = 0; i < parameters.length; i++)
			{
				parameterObjects[i] = this.evaluate(parameters[i].getValueExpression(), contextClassifier, contextInstance, variablesScope)
						.getValue();

				parameterTypes[i] = parameters[i].getType().getInstanceClass();

				if (parameterTypes[i] == null)
				{
					throw new SDMException("Instance class of '" + parameters[i].getType() + "' is null.");
				}
			}

			// Get class and method
			Class<?> c;
			Method m;

			Object methodInstanceObject = null;

			// static method call
			if (callAction.getThisParameterValue() == null)
			{
				c = Class.forName(callAction.getMethodClassName(), true, this.getExpressionInterpreterManager().getClassLoader());
				m = c.getMethod(callAction.getMethodName(), parameterTypes);
			}
			else
			{
				// dynamic method call

				methodInstanceObject = this
						.evaluate(callAction.getThisParameterValue(), contextClassifier, contextInstance, variablesScope).getValue();

				if (methodInstanceObject != null)
				{
					c = methodInstanceObject.getClass();
				}
				else
				{
					throw new NullPointerException("methodInstanceObject expression '" + callAction.getThisParameterValue()
							+ "'evaluated to null");
				}

				m = c.getMethod(callAction.getMethodName(), parameterTypes);
			}

			// Invoke method
			final Object returnValue = m.invoke(methodInstanceObject, parameterObjects);

			if (callAction.getType() == null)
			{
				/*
				 * A method without return value
				 */
				if (returnValue != null)
				{
					throw new SDMException("MethodCallAction '" + callAction + "' has no classifier but the called method returned '"
							+ returnValue + "'.");
				}

				return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EJAVA_OBJECT, null);
			}
			else
			{
				return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getType(), returnValue);
			}

		}
		catch (final Exception e)
		{
			throw new SDMException("Error executing method call action: " + callAction, e);
		}
	}

	protected Variable<EClassifier> evaluate(final EOperationCallAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		/*
		 * Get this variable
		 */
		final Variable<EClassifier> thisVariable = this.evaluate(callAction.getThisParameterValue(), contextClassifier, contextInstance,
				variablesScope);

		if (thisVariable.getClassifier().eClass() != EcorePackage.Literals.ECLASS)
		{
			throw new SDMException("An EOperation can only be invoked on an EObject, not on '" + thisVariable.getValue() + "'.");
		}

		/*
		 * Compile list of arguments
		 */
		final EList<Object> arguments = new BasicEList<Object>();

		for (final CallActionParameter parameter : callAction.getParameters())
		{
			arguments.add(this.evaluate(parameter.getValueExpression(), contextClassifier, contextInstance, variablesScope).getValue());
		}

		/*
		 * Invoke the operation
		 */
		try
		{
			final Object returnValue = ((EObject) thisVariable.getValue()).eInvoke(callAction.getEOperation(), arguments);

			return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getType(), returnValue);
		}
		catch (final InvocationTargetException ex)
		{
			throw new SDMException("Could not invoke operation '" + callAction.getEOperation() + "' on '" + thisVariable.getValue() + "'.",
					ex);
		}
	}

	protected Variable<EClassifier> evaluate(final NullValueAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
	{
		return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getType(), null);
	}

	protected Variable<EClassifier> evaluate(final UnaryOperationAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		final Variable<EClassifier> v1 = this.evaluate(callAction.getOperand(), contextClassifier, contextInstance, variablesScope);
		final Object o1 = v1.getValue();

		switch (callAction.getOperator())
		{
			case NOT:
				if (o1 instanceof Boolean)
				{
					return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EBOOLEAN,
							!(Boolean) o1);
				}
				else
				{
					throw new SDMException("Error executing operation action: " + callAction
							+ " (Logical operators may only be applied to type Boolean)");
				}
			default:
				throw new UnsupportedOperationException();
		}
	}

	protected Variable<EClassifier> evaluate(final BinaryOperationAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		final Variable<EClassifier> v1 = this.evaluate(callAction.getOperand1(), contextClassifier, contextInstance, variablesScope);
		final Object o1 = v1.getValue();

		final Variable<EClassifier> v2 = this.evaluate(callAction.getOperand2(), contextClassifier, contextInstance, variablesScope);
		final Object o2 = v2.getValue();

		switch (callAction.getOperator())
		{
			case EQUALS:
			case NOT_EQUAL:
			case GREATER_OR_EQUALS:
			case GREATER_THAN:
			case LESS_OR_EQUALS:
			case LESS_THAN:
			{
				boolean value;
				if (o1 == null || o2 == null)
				{
					value = o1 == o2;
				}
				else if (o1 instanceof Comparable<?> && o2 instanceof Comparable<?> && o1.getClass().equals(o2.getClass()))
				{
					@SuppressWarnings(
					{
							"unchecked", "rawtypes"
					})
					final int result = ((Comparable) o1).compareTo(o2);
					switch (callAction.getOperator())
					{
						case EQUALS:
							value = result == 0 ? true : false;
							break;
						case GREATER_OR_EQUALS:
							value = result >= 0 ? true : false;
							break;
						case GREATER_THAN:
							value = result > 0 ? true : false;
							break;
						case LESS_OR_EQUALS:
							value = result <= 0 ? true : false;
							break;
						case LESS_THAN:
							value = result < 0 ? true : false;
							break;
						case NOT_EQUAL:
							value = result != 0 ? true : false;
							break;
						default:
							throw new UnsupportedOperationException();
					}
				}
				else if (o1 instanceof Number && o2 instanceof Number)
				{
					switch (callAction.getOperator())
					{
						case EQUALS:
							value = ((Number) o1).doubleValue() == ((Number) o2).doubleValue();
							break;
						case GREATER_OR_EQUALS:
							value = ((Number) o1).doubleValue() >= ((Number) o2).doubleValue();
							break;
						case GREATER_THAN:
							value = ((Number) o1).doubleValue() > ((Number) o2).doubleValue();
							break;
						case LESS_OR_EQUALS:
							value = ((Number) o1).doubleValue() <= ((Number) o2).doubleValue();
							break;
						case LESS_THAN:
							value = ((Number) o1).doubleValue() < ((Number) o2).doubleValue();
							break;
						case NOT_EQUAL:
							value = ((Number) o1).doubleValue() != ((Number) o2).doubleValue();
							break;
						default:
							throw new UnsupportedOperationException();
					}
				}
				else if (callAction.getOperator() == BinaryOperatorsEnum.EQUALS)
				{
					value = o1 == o2;
				}
				else if (callAction.getOperator() == BinaryOperatorsEnum.NOT_EQUAL)
				{
					value = o1 != o2;
				}
				else
				{
					throw new SDMException(
							"Error executing operation action: "
									+ callAction
									+ " (If objects need to be compared, they must implement java.lang.Comparable, or only '=' and '!=' comparisons are possible.)");
				}

				return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EBOOLEAN, value);
			}
			case AND:
			case OR:
			{
				if (o1 instanceof Boolean && o2 instanceof Boolean)
				{
					boolean value;
					final Boolean b1 = (Boolean) o1;
					final Boolean b2 = (Boolean) o2;

					switch (callAction.getOperator())
					{
						case AND:
							value = b1 && b2;
							break;
						case OR:
							value = b1 || b2;
							break;
						default:
							throw new UnsupportedOperationException();
					}

					return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EBOOLEAN, value);
				}
				else
				{
					throw new SDMException("Error executing operation action: " + callAction
							+ " (Logical operators may only be applied to type Boolean)");
				}
			}

			case ADD:
				if (o1 instanceof String && o2 instanceof String)
				{
					return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.ESTRING, (String) o1
							+ (String) o2);
				}
			case DIVIDE:
			case MODULO:
			case MULTIPLY:
			case SUBTRACT:
			{
				if (o1 instanceof Number && o2 instanceof Number)
				{

					final BigDecimal bd1 = new BigDecimal(o1.toString());
					final BigDecimal bd2 = new BigDecimal(o2.toString());

					BigDecimal result;
					switch (callAction.getOperator())
					{
						case ADD:
							result = bd1.add(bd2);
							break;
						case SUBTRACT:
							result = bd1.subtract(bd2);
							break;
						case MULTIPLY:
							result = bd1.multiply(bd2);
							break;
						case DIVIDE:
							result = bd1.divide(bd2, RoundingMode.HALF_UP);
							break;
						case MODULO:
							result = bd1.remainder(bd2);
							break;
						default:
							throw new UnsupportedOperationException();
					}

					final EClassifier classifier = callAction.getType();

					Object finalResult;

					if (classifier == EcorePackage.Literals.EBYTE || classifier == EcorePackage.Literals.EBYTE_OBJECT)
					{
						finalResult = result.byteValue();
					}
					else if (classifier == EcorePackage.Literals.EINT || classifier == EcorePackage.Literals.EINTEGER_OBJECT)
					{
						finalResult = result.intValue();
					}
					else if (classifier == EcorePackage.Literals.EDOUBLE || classifier == EcorePackage.Literals.EDOUBLE_OBJECT)
					{
						finalResult = result.doubleValue();
					}
					else if (classifier == EcorePackage.Literals.EFLOAT || classifier == EcorePackage.Literals.EFLOAT_OBJECT)
					{
						finalResult = result.floatValue();
					}
					else if (classifier == EcorePackage.Literals.ELONG || classifier == EcorePackage.Literals.ELONG_OBJECT)
					{
						finalResult = result.longValue();
					}
					else if (classifier == EcorePackage.Literals.ESHORT || classifier == EcorePackage.Literals.ESHORT_OBJECT)
					{
						finalResult = result.shortValue();
					}
					else
					{
						throw new UnsupportedOperationException();
					}

					return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, classifier, finalResult);
				}
				else
				{
					throw new UnsupportedOperationException();
				}
			}
			default:
				throw new UnsupportedOperationException();
		}
	}

	protected Variable<EClassifier> evaluate(final VariableDeclarationAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		assert callAction.getValueExpression() != null : "The value expression is null in " + callAction + ".";

		// Evaluate the assignment;
		final Object value = this.evaluate(callAction.getValueExpression(), contextClassifier, contextInstance, variablesScope).getValue();

		/*
		 * Check that the types are compatible. Null is considered compatible to
		 * all types.
		 */
		if (value != null && !callAction.getType().isInstance(value))
		{
			throw new SDMException(value + " is not an instance of " + callAction.getType().getName());
		}

		final Variable<EClassifier> variable = variablesScope.createVariable(callAction.getName(), callAction.getType(), value);

		return variable;
	}

	protected Variable<EClassifier> evaluate(final LiteralDeclarationAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		final LiteralDeclarationAction literalDeclarationAction = callAction;

		/*
		 * Let EcoreFactory create an Object of the appropriate primitive type
		 */
		if (literalDeclarationAction.getType() instanceof EDataType)
		{
			final EDataType eDataType = (EDataType) literalDeclarationAction.getType();

			return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, eDataType, eDataType.getEPackage()
					.getEFactoryInstance().createFromString(eDataType, literalDeclarationAction.getLiteral()));
		}
		else
		{
			throw new SDMException(literalDeclarationAction.getType() + " must be an EDataType.");
		}
	}
	
	@SuppressWarnings("unchecked")
	protected MLSDMInterpreter createMLSDMInterpreter() {
		return new MLSDMInterpreter( (ExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.getExpressionInterpreterManager(),
				(NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this.getExpressionInterpreterManager().getNotificationEmitter(),
				MLSDMMetamodelFacadeFactory.INSTANCE );
	}

	protected Variable<EClassifier> evaluate(final ActivityCallAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException	{
		final MLSDMInterpreter sdi = createMLSDMInterpreter();
//		@SuppressWarnings("unchecked")
//		final MLSDMInterpreter sdi = new MLSDMInterpreter(
//				(ExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this
//						.getExpressionInterpreterManager(),
//				(NotificationEmitter<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression>) this
//						.getExpressionInterpreterManager().getNotificationEmitter() );

		/*
		 * Create list of variable bindings.
		 */
		final List<Variable<EClassifier>> parameters = new LinkedList<Variable<EClassifier>>();

		for (final CallActionParameter parameter : callAction.getParameters())
		{
			final Variable<EClassifier> variable = new Variable<EClassifier>(parameter.getName(), parameter.getType(), this.evaluate(
					parameter.getValueExpression(), contextClassifier, contextInstance, variablesScope).getValue());

			parameters.add(variable);
		}

		final Map<String, Variable<EClassifier>> returnValues = sdi.executeActivity(callAction.getActivity(), parameters);

		Variable<EClassifier> returnVariable = null;

		if (callAction.getReturnValueParameter() != null)
		{
			returnVariable = returnValues.get(callAction.getReturnValueParameter().getName());
			returnVariable = new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, returnVariable.getClassifier(),
					returnVariable.getValue());
		}
		else
		{
			returnVariable = new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, EcorePackage.Literals.EJAVA_OBJECT, null);
		}

		return returnVariable;
	}

	protected Variable<EClassifier> evaluate(final NewObjectAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		if (callAction.getType().eClass() == EcorePackage.Literals.ECLASS)
		{
			assert callAction.getParameters().isEmpty();

			return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getType(), callAction.getType()
					.getEPackage().getEFactoryInstance().create((EClass) callAction.getType()));
		}
		else if (callAction.getType().eClass() == EcorePackage.Literals.EDATA_TYPE)
		{
			try
			{
				final CallActionParameter[] parameters = callAction.getParameters().toArray(new CallActionParameter[] {});

				// Create array of parameter types and parameters
				final Class<?>[] parameterTypes = new Class<?>[parameters.length];
				final Object[] parameterObjects = new Object[parameters.length];

				for (int i = 0; i < parameters.length; i++)
				{
					parameterObjects[i] = this.evaluate(parameters[i].getValueExpression(), contextClassifier, contextInstance,
							variablesScope).getValue();

					parameterTypes[i] = parameters[i].getType().getInstanceClass();
				}

				final Class<?> c = Class.forName(callAction.getType().getInstanceTypeName(), true, this.getExpressionInterpreterManager()
						.getClassLoader());

				final Constructor<?> constructor = c.getConstructor(parameterTypes);

				return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getType(),
						constructor.newInstance(parameterObjects));
			}
			catch (final Exception ex)
			{
				throw new SDMException("Error while creating new object: " + callAction, ex);
			}
		}
		else
		{
			throw new UnsupportedOperationException();
		}
	}

	protected Variable<EClassifier> evaluate(final CloneAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
			throws SDMException
	{
		assert callAction.getValueExpression() != null;

		final Variable<EClassifier> variable = this.evaluate(callAction.getValueExpression(), contextClassifier, contextInstance,
				variablesScope);

		/*
		 * If the object is an EObject, use EcoreUtil.copy() to clone it.
		 */
		if (variable.getClassifier().eClass() == EcorePackage.Literals.ECLASS)
		{
			return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, variable.getClassifier(),
					EcoreUtil.copy((EObject) variable.getValue()));
		}
		else
		{
			throw new SDMException("The value '" + variable.getValue() + "' is not an EObject.");
		}
	}

	protected Variable<EClassifier> evaluate(final EObjectReferenceAction callAction, final EClassifier contextClassifier,
			final Object contextInstance, final VariablesScope<?, ?, ?, ?, ?, ?, EClassifier, ?, MLExpression> variablesScope)
	{
		assert callAction.getEObject() != null;

		return new Variable<EClassifier>(SDMInterpreterConstants.INTERNAL_VAR_NAME, callAction.getEObject().eClass(),
				callAction.getEObject());
	}

}
